﻿/* 
 * Nan.Contests.MvcTriangleContest-2
 * Copyright © 2014 Julio P. Di Egidio <julio@diegidio.name>
 * This program is released into the public domain.
 * No warranty of any kind is implied.
 */

using System;
using System.ComponentModel;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Numerics;
using System.Threading;
using System.Windows.Forms;

namespace Nan.Contests.MvcTriangleContest_2.Display
{

	public partial class Form1 : Form
	{

		public readonly static string FormName = typeof(Form1).Assembly.GetName().Name;

		private int _mainTriangleSide = 0;
		private int _innerTriangleSide = 0;
		private int _drawingOffset = 0;
		private bool _wireframe = true;
		private bool _autoUpdate = true;
		private SizeF _graphicsTrianglesDownSize;
		private SizeF _graphicsTrianglesUpSize;
		private Graphics _graphicsTrianglesDown;
		private Graphics _graphicsTrianglesUp;
		private BackgroundWorker _worker;

		public Form1()
		{
			InitializeComponent();

			Text = FormName;

			_graphicsTrianglesDownSize = new SizeF(
				pbxTrianglesDown.DisplayRectangle.Width - (pbxTrianglesDown.BorderStyle == BorderStyle.None ? 0.0f : 2.0f),
				pbxTrianglesDown.DisplayRectangle.Height - (pbxTrianglesDown.BorderStyle == BorderStyle.None ? 0.0f : 2.0f)
			);

			_graphicsTrianglesUpSize = new SizeF(
				pbxTrianglesUp.DisplayRectangle.Width - (pbxTrianglesUp.BorderStyle == BorderStyle.None ? 0.0f : 2.0f),
				pbxTrianglesUp.DisplayRectangle.Height - (pbxTrianglesUp.BorderStyle == BorderStyle.None ? 0.0f : 2.0f)
			);

			_graphicsTrianglesDown = GetGraphics(pbxTrianglesDown, _graphicsTrianglesDownSize.Width);
			_graphicsTrianglesUp = GetGraphics(pbxTrianglesUp, _graphicsTrianglesUpSize.Width);

			_worker = GetWorker();

			nudMainTriangleSide.Value = _mainTriangleSide;
			nudInnerTriangleSide.Value = _innerTriangleSide;
			nudDrawingOffset.Value = _drawingOffset;
			chkWireframe.Checked = _wireframe;
			chkAutoUpdate.Checked = _autoUpdate;

			nudMainTriangleSide.ValueChanged += nudMainTriangleSide_ValueChanged;
			nudInnerTriangleSide.ValueChanged += nudInnerTriangleSide_ValueChanged;
			nudDrawingOffset.ValueChanged += nudDrawingOffset_ValueChanged;
			chkWireframe.CheckedChanged += chkWireframe_CheckedChanged;
			chkAutoUpdate.CheckedChanged += chkAutoUpdate_CheckedChanged;
			btnUpdate.Click += btnUpdate_Click;

			StartWork();
		}

		private readonly static object _sharedLock = new object();

		private void nudMainTriangleSide_ValueChanged(object sender, EventArgs e)
		{
			lock (_sharedLock)
			{
				var value = (int)nudMainTriangleSide.Value;

				if (value != _mainTriangleSide)
				{
					CancelWork();

					_mainTriangleSide = value;

					if (_mainTriangleSide < _innerTriangleSide)
					{
						_innerTriangleSide = _mainTriangleSide;

						nudInnerTriangleSide.Value = _innerTriangleSide;
					}

					if (_autoUpdate)
					{
						StartWork();
					}
				}
			}
		}

		private void nudInnerTriangleSide_ValueChanged(object sender, EventArgs e)
		{
			lock (_sharedLock)
			{
				var value = (int)nudInnerTriangleSide.Value;

				if (value != _innerTriangleSide)
				{
					if (value > _mainTriangleSide)
					{
						nudInnerTriangleSide.Value = _mainTriangleSide;

						return;
					}

					CancelWork();

					_innerTriangleSide = value;

					if (_autoUpdate)
					{
						StartWork();
					}
				}
			}
		}

		private void nudDrawingOffset_ValueChanged(object sender, EventArgs e)
		{
			lock (_sharedLock)
			{
				CancelWork();

				_drawingOffset = (int)nudDrawingOffset.Value;

				if (_autoUpdate)
				{
					StartWork();
				}
			}
		}

		private void chkWireframe_CheckedChanged(object sender, EventArgs e)
		{
			lock (_sharedLock)
			{
				CancelWork();

				_wireframe = chkWireframe.Checked;

				if (_autoUpdate)
				{
					StartWork();
				}
			}
		}

		private void chkAutoUpdate_CheckedChanged(object sender, EventArgs e)
		{
			lock (_sharedLock)
			{
				_autoUpdate = chkAutoUpdate.Checked;

				if (_autoUpdate)
				{
					CancelWork();

					StartWork();
				}
			}
		}

		private void btnUpdate_Click(object sender, EventArgs e)
		{
			lock (_sharedLock)
			{
				CancelWork();

				StartWork();
			}
		}

		private readonly static float _sqrt2 = (float)Math.Sqrt(2.0);

		private Graphics GetGraphics(Control owner, float width)
		{
			var graphics = owner.CreateGraphics();

			graphics.CompositingMode = CompositingMode.SourceOver;
			graphics.CompositingQuality = CompositingQuality.HighQuality;
			graphics.PageScale = 1.0f;
			graphics.PageUnit = GraphicsUnit.Pixel;
			graphics.PixelOffsetMode = PixelOffsetMode.None;
			graphics.RenderingOrigin = new Point(0, 0);
			graphics.SmoothingMode = SmoothingMode.None;

			var transform = new Matrix();

			transform.Translate(width / 2.0f, 0.0f);
			transform.Scale(1.0f / _sqrt2, 1.0f * _sqrt2);
			transform.RotateAt(45.0f, new PointF(0.0f, 0.0f));

			graphics.Transform = transform;

			return graphics;
		}

		private BackgroundWorker GetWorker()
		{
			var worker = new BackgroundWorker()
			{
				WorkerSupportsCancellation = true
			};

			worker.DoWork += _worker_DoWork;
			worker.RunWorkerCompleted += _worker_RunWorkerCompleted;

			return worker;
		}

		private void CancelWork()
		{
			if (_worker.IsBusy)
			{
				if (!_worker.CancellationPending)
				{
					_worker.CancelAsync();
				}

				while (_worker.IsBusy)
				{
					Application.DoEvents();

					Thread.Sleep(0);
				}
			}
		}

		private void StartWork()
		{
			Cursor = Cursors.AppStarting;

			ClearDisplay();

			SetCounts();

			_worker.RunWorkerAsync();
		}

		private void _worker_DoWork(object sender, DoWorkEventArgs e)
		{
			DoWork();

			if (_worker.CancellationPending)
			{
				e.Cancel = true;
			}
		}

		private void _worker_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e)
		{
			Cursor = Cursors.Default;
		}

		private void DoWork()
		{
			foreach (var triangle in Solver.GetTriangles(_mainTriangleSide, _innerTriangleSide))
			{
				if (_worker.CancellationPending)
				{
					return;
				}

				Invoke(new Action(() =>
				{
					try
					{
						DrawTriangle(triangle);
					}
					catch { }
				}));
			}
		}

		private const float _penWidth = 1.0f;
		private const float _offsetMulti = 10.0f;

		private void ClearDisplay()
		{
			_graphicsTrianglesDown.Clear(pbxTrianglesDown.BackColor);
			_graphicsTrianglesUp.Clear(pbxTrianglesUp.BackColor);
		}

		private void SetCounts()
		{
			BigInteger downCnt = Solver.CountTrianglesDown(_mainTriangleSide, _innerTriangleSide);
			BigInteger upCnt = Solver.CountTrianglesUp(_mainTriangleSide, _innerTriangleSide);
			BigInteger cnt = Solver.CountTriangles(_mainTriangleSide, _innerTriangleSide);

			txtCountTrianglesDown.Text = downCnt.ToString();
			txtCountTrianglesUp.Text = upCnt.ToString();
			txtCountTriangles.Text = cnt.ToString();
		}

		private void DrawTriangle(Triangle triangle)
		{
			SizeF graphicsSize;
			Graphics graphics;

			int row, col, hei, side;

			if (triangle.IsDown)
			{
				graphicsSize = _graphicsTrianglesDownSize;
				graphics = _graphicsTrianglesDown;

				row = triangle.Center.Row;
				col = triangle.Center.Col;
				hei = triangle.Height;

				side = _mainTriangleSide;
			}
			else
			{
				graphicsSize = _graphicsTrianglesUpSize;
				graphics = _graphicsTrianglesUp;

				row = triangle.Left.Row;
				col = triangle.Left.Col;
				hei = -triangle.Height;

				side = _mainTriangleSide;
			}

			float ratR = side == 1 ? 0.0f : (float)row / (side - 1);
			float ratR1 = side == 0 ? 0.0f : (float)(row + 1) / side;
			float ratC = row == 0 ? 0.0f : (float)col / row;
			float ratH = side == 0 ? 0.0f : (float)hei / side;

			float basB = 1.0f - ratR;
			float basR = ratC;
			float basG = (1.0f - ratC) * ratR;

			float colB = basB + (1.0f - basB) * ratH;
			float colR = basR + (1.0f - basR) * ratH;
			float colG = basG + (1.0f - basG) * ratH;

			var blue = (int)(255.0f * colB);
			var red = (int)(255.0f * colR);
			var green = (int)(255.0f * colG);

			float offset = _drawingOffset * _offsetMulti * (ratH * ratR1);

			PointF[] points;

			if (triangle.IsDown)
			{
				points = new PointF[]
				{
 					CoordToPointF(triangle.Center, graphicsSize, offset),
 					CoordToPointF(triangle.Left, graphicsSize),
 					CoordToPointF(triangle.Right, graphicsSize),
				};
			}
			else
			{
				points = new PointF[]
				{
 					CoordToPointF(triangle.Center, graphicsSize),
 					CoordToPointF(triangle.Left, graphicsSize, offset),
 					CoordToPointF(triangle.Right, graphicsSize, offset),
				};
			}

			var color = Color.FromArgb(red, green, blue);

			if (_wireframe)
			{
				var pen = new Pen(color, _penWidth);

				graphics.DrawPolygon(pen, points);
			}
			else
			{
				var brush = new SolidBrush(color);

				graphics.FillPolygon(brush, points);
			}
		}

		private PointF CoordToPointF(Coord coord, SizeF graphicsSize, float offset = 0.0f)
		{
			var c = (float)coord.Col;
			var r = (float)(coord.Row - coord.Col);
			var s = (float)_mainTriangleSide;

			var x = graphicsSize.Width * c / s + offset;
			var y = graphicsSize.Height * r / s + offset;

			return new PointF(x, y);
		}

	}

}
