﻿/* 
 * Nan.Contests.MvcTriangleContest-2
 * Copyright © 2014 Julio P. Di Egidio <julio@diegidio.name>
 * This program is released into the public domain.
 * No warranty of any kind is implied.
 */

using System;
using System.Collections.Generic;
using System.Numerics;

namespace Nan.Contests.MvcTriangleContest_2
{

	public static class Solver
	{

		#region Interface

		private static void checkSide(int side, string paramName)
		{
			if (side < 0)
			{
				throw new ArgumentOutOfRangeException(paramName, side, "Value must be non-negative.");
			}
		}

		public static BigInteger CountTrianglesDown(int mainSide, int innerSide = 0)
		{
			checkSide(mainSide, "mainSide");
			checkSide(innerSide, "innerSide");

			if (innerSide == 0)
			{
				return CountTrianglesDown(mainSide);
			}
			else
			{
				return CountTrianglesDownOfSide(mainSide, innerSide);
			}
		}

		public static BigInteger CountTrianglesUp(int mainSide, int innerSide = 0)
		{
			checkSide(mainSide, "mainSide");
			checkSide(innerSide, "innerSide");

			if (innerSide == 0)
			{
				return CountTrianglesUp(mainSide);
			}
			else
			{
				return CountTrianglesUpOfSide(mainSide, innerSide);
			}
		}

		public static BigInteger CountTriangles(int mainSide, int innerSide = 0)
		{
			checkSide(mainSide, "mainSide");
			checkSide(innerSide, "innerSide");

			if (innerSide == 0)
			{
				return CountTriangles(mainSide);
			}
			else
			{
				return CountTrianglesByOfSide(mainSide, innerSide);
			}
		}

		public static IEnumerable<Triangle> GetTriangles(int mainSide, int innerSide = 0)
		{
			checkSide(mainSide, "mainSide");
			checkSide(innerSide, "innerSide");

			if (innerSide == 0)
			{
				return GetTrianglesAllSides(mainSide);
			}
			else
			{
				return GetTrianglesOfSide(mainSide, innerSide);
			}
		}

		#endregion

		#region Implementation

		private static BigInteger CountTrianglesDown(int mainSide)
		{
			// Number of triangles with same orientation as largest triangle.
			// https://oeis.org/A000292
			// C_d(n) := n(n+1)(n+2)/6

			var n = new BigInteger(mainSide);

			return n * (n + 1) * (n + 2) / 6;
		}

		private static BigInteger CountTrianglesUp(int mainSide)
		{
			// Number of triangles pointing in opposite direction to largest triangle.
			// https://oeis.org/A002623
			// C_u(n+2) := FLOOR((n+2)(n+4)(2n+3)/24) 

			var n = new BigInteger(mainSide);

			return n * (n + 2) * (2 * n - 1) / 24;
		}

		private static BigInteger CountTriangles(int mainSide)
		{
			// Number of triangles in triangular matchstick arrangement of side n.
			// https://oeis.org/A002717
			// C(n) := FLOOR(n(n+2)(2n+1)/8)

			var n = new BigInteger(mainSide);

			return n * (n + 2) * (2 * n + 1) / 8;
		}

		private static BigInteger CountTrianglesDownOfSide(int mainSide, int innerSide)
		{
			// K_d(n, m) := n-m+1
			// C_d(n, m) := K_d(n, m)(K_d(n, m)+1)/2

			var n = new BigInteger(mainSide);
			var m = new BigInteger(innerSide);

			var k = n - m + 1;

			return k * (k + 1) / 2;
		}

		private static BigInteger CountTrianglesUpOfSide(int mainSide, int innerSide)
		{
			// K_u(n, m) := n < 2m-1 ? 0 : n-(2m-1)
			// C_u(n, m) := K_u(n, m)(K_u(n, m)+1)/2

			var n = new BigInteger(mainSide);
			var m = new BigInteger(innerSide);

			var _2m_1 = 2 * m - 1;

			var k = n < _2m_1 ? BigInteger.Zero : n - _2m_1;

			return k * (k + 1) / 2;
		}

		private static BigInteger CountTrianglesByOfSide(int mainSide, int innerSide)
		{
			// C(n, m) := C_d(n, m) + C_u(n, m) =
			//          = K_d(n, m)(K_d(n, m)+1)/2 + K_u(n, m)(K_u(n, m)+1)/2 =
			//          = (n-m+1)(n-m+2)/2 + (n < 2m-1 ? 0 : (n-2m+1)(n-2m+2)/2) =
			//          = n < 2m-1 ? (n-m+1)(n-m+2)/2 : (n^2-nm+2n-nm+m^2-2m+n-m+2 + n^2-2nm+2n-2nm+4m^2-4m+n-2m+2)/2 =
			//          = n < 2m-1 ? (n-m+1)(n-m+2)/2 : (2n^2-6nm+6n+5m^2-9m+4)/2

			var n = new BigInteger(mainSide);
			var m = new BigInteger(innerSide);

			var _2m_1 = 2 * m - 1;

			if (n < _2m_1)
			{
				return CountTrianglesDownOfSide(mainSide, innerSide);
			}

			return (2 * (n * n) - 6 * n * m + 6 * n + 5 * (m * m) - 9 * m + 4) / 2;
		}

		public static IEnumerable<Triangle> GetTrianglesAllSides(int mainSide)
		{
			for (var cRow = 0; cRow <= mainSide; ++cRow)
			{
				for (var cCol = 0; cCol <= cRow; ++cCol)
				{
					var center = new Coord(cRow, cCol);

					// GetTrianglesDown
					{
						var dCol = cCol + 1;

						for (var dRow = cRow + 1; dRow <= mainSide; ++dRow)
						{
							var left = new Coord(dRow, cCol);
							var right = new Coord(dRow, dCol);

							yield return new Triangle(center, left, right);

							++dCol;
						}
					}

					// GetTrianglesUp
					{
						var lRow = Math.Max(cCol, cRow - cCol);

						var dCol = cCol - 1;

						for (var dRow = cRow - 1; dRow >= lRow; --dRow)
						{
							var left = new Coord(dRow, dCol);
							var right = new Coord(dRow, cCol);

							yield return new Triangle(center, left, right);

							--dCol;
						}
					}
				}
			}
		}

		public static IEnumerable<Triangle> GetTrianglesOfSide(int mainSide, int innerSide)
		{
			for (var cRow = 0; cRow <= mainSide; ++cRow)
			{
				for (var cCol = 0; cCol <= cRow; ++cCol)
				{
					// GetTrianglesDown
					{
						var dRow = cRow + innerSide;

						if (dRow <= mainSide)
						{
							var center = new Coord(cRow, cCol);

							var dCol = cCol + innerSide;

							var left = new Coord(dRow, cCol);
							var right = new Coord(dRow, dCol);

							yield return new Triangle(center, left, right);
						}
					}

					// GetTrianglesUp
					{
						var lRow = Math.Max(cCol, cRow - cCol);

						var dRow = cRow - innerSide;

						if (dRow >= lRow)
						{
							var center = new Coord(cRow, cCol);

							var dCol = cCol - innerSide;

							var left = new Coord(dRow, dCol);
							var right = new Coord(dRow, cCol);

							yield return new Triangle(center, left, right);
						}
					}
				}
			}
		}

		#endregion

	}

}
